%% Script for trying some models on coupled EEM NMR LC-MS data with Tensorlab
%--------------Cleaning the workspace and loading functions---------------%
clc;
clear     all;
close     all;
%-------------------------------------------------------------------------%


% Flexible coupling and compression
% ---------------------------------
%
% File created for Jeremy Cohen in reply to an email on August 16, 2016.

% Author(s): Nico Vervliet
%
% Version History:
% - 2016/08/16   NV      Initial version


%load('/Users/cohenjer/Documents/Travail/MATLAB/Data/Data Chemo/data/Multimodal/EEM_NMR_LCMS.mat')
load('D:/Travail/Travail/MATLAB/Data/Data Chemo/data/Multimodal/EEM_NMR_LCMS.mat')

%------------------------ Missing values preprocessing -------------------%

Tx        =    X.data;
Ty        =    Y.data;

Tx        =    permute(Tx,[3,2,1]);
Ty        =    permute(Ty,[3,2,1]);

dim_X     =     size(Tx);
dim_Y     =     size(Ty);



Tx        =    reshape(Tx,prod(dim_X),1);
Ty        =    reshape(Ty,prod(dim_Y),1);

temp_x         =     Tx(isnan(Tx));
temp_y         =     Ty(isnan(Ty));

Tx(isnan(Tx))  =     0;%abs(0.01*randn(size(temp_x)));
Ty(isnan(Ty))  =     0;%abs(0.01*randn(size(temp_y)));

Tx             =     reshape(Tx,dim_X);
Ty             =     reshape(Ty,dim_Y);

normTx         =     (Tx(:))'*(Tx(:));
normTy         =     (Ty(:))'*(Ty(:));
Tx             =     Tx/sqrt(normTx); 
Ty             =     Ty/sqrt(normTy);

size_one = dim_X;
size_two = dim_Y;
R = 3;
R2 = 5;
snr1 = 40; % Signal to noise ratio for tensor 1 in dB
snr2 = 40; % Signal to noise ratio for tensor 2 in dB
snrC = 40; % C1 to C2 noise ratio dB

display('Compression of the tensors');
tic
[Uc1, Sc1] = mlsvd_rsi(Tx, [6 6 6]);
toc
tic
[Uc2, Sc2] = mlsvd_rsi(Ty, [6 6 6]);
toc



%%
disp('** Full tensors - hard coupling')

model = struct;
model.variables = cpd_rnd([size_one, size_two(1:2)], R, 'real', @rand);
model.transform = {1:5, @struct_nonneg};
model.factorizations{1}.data = Tx;
model.factorizations{1}.cpd = 1:3;
model.factorizations{1}.weight = 1/snr1;
model.factorizations{2}.data = Ty;
model.factorizations{2}.cpd = [4 5 3];
model.factorizations{2}.weight = 1/snr2;

tic
[sol1, out1] = sdf_nls(model, 'Display', 10, 'CGMaxIter', 100,'TolFun', 1e-10);
toc



%%
disp('** Compressed tensors - hard coupling')

model = struct;
model.variables = cpd_rnd([size_one, size_two(1:2)], R, 'real', @rand);
model.transform = {1:5, @struct_nonneg};
model.factorizations{1}.data = {Uc1, Sc1};
model.factorizations{1}.cpd = 1:3;
model.factorizations{1}.weight = 1/snr1;
model.factorizations{2}.data = {Uc2, Sc2};
model.factorizations{2}.cpd = [4 5 3];
model.factorizations{2}.weight = 1/snr2;

tic
[sol2, out2] = sdf_nls(model, 'Display', 10, 'CGMaxIter', 100,'TolFun', 1e-10);
toc


%% 
disp('** Compressed tensors - no coupling')

model = struct;
model.variables = cpd_rnd([size_one, size_two], R, 'real', @rand);
model.transform = {1:6, @struct_nonneg};
model.factorizations{1}.data = {Uc1, Sc1};
model.factorizations{1}.cpd = 1:3;
model.factorizations{1}.weight = 1/snr1;
model.factorizations{2}.data = {Uc2, Sc2};
model.factorizations{2}.cpd = 4:6;
model.factorizations{2}.weight = 1/snr2;
tic
[sol3, out3] = sdf_nls(model, 'Display', 100, 'CGMaxIter', 100, 'TolFun', 1e-11);
toc


%%
disp('Compressed tensors - flexible coupling')

model = struct;
model.variables = cpd_rnd([size_one, size_two], R, 'real', @rand);
model.transform = {1:6, @struct_nonneg};
model.factors{8} = [eye(R), -eye(R)];
model.factors{7} = {3, 6};
model.factorizations{1}.data = {Uc1, Sc1};
model.factorizations{1}.cpd = 1:3;
model.factorizations{1}.weight = 1/snr1;
model.factorizations{2}.data = {Uc2, Sc2};
model.factorizations{2}.cpd = 4:6;
model.factorizations{2}.weight = 1/snr2;
model.factorizations{3}.data = zeros(size_one(3),R);
model.factorizations{3}.cpd = 7:8;
model.factorizations{3}.weight = 1/snrC;
tic
[sol4, out4] = sdf_nls(model, 'Display', 10, 'CGMaxIter', 100, 'TolFun', 1e-11);
toc


%%
disp('Compressed tensors - flexible partial coupling')

model = struct;
model.variables.a = rand(size_one(1),R);
model.variables.b = rand(size_one(2),R);
model.variables.c1 = rand(size_one(3),R);
model.variables.d = rand(size_two(1),R2);
model.variables.e = rand(size_two(2),R2);
model.variables.c21 = rand(size_two(3),R);
model.variables.c22 = rand(size_two(3),R2-R);
model.factors.A = {'a', @struct_nonneg};
model.factors.B = {'b', @struct_nonneg};
model.factors.D = {'d', @struct_nonneg};
model.factors.E = {'e', @struct_nonneg};
model.factors.C1 = {'c1', @struct_nonneg}; % use in factorization of T1
model.factors.C2 = {{'c21', @struct_nonneg}, {'c22', @struct_nonneg}}; % use in factorization of T2
model.factors.C12 = {{'c1', @struct_nonneg}, {'c21', @struct_nonneg}}; % use in flexible coupling term c1 approx c21
model.factors.Coupl = [eye(R), -eye(R)];
model.factorizations{1}.data = {Uc1, Sc1};
model.factorizations{1}.cpd ={'A','B','C1'};
model.factorizations{1}.weight = 1/snr1;
model.factorizations{2}.data = {Uc2, Sc2};
model.factorizations{2}.cpd = {'D','E','C2'};
model.factorizations{2}.weight = 1/snr2;
model.factorizations{3}.data = zeros(size_one(3),R);
model.factorizations{3}.cpd = {'C12','Coupl'};
model.factorizations{3}.weight = 1/snrC;
tic
[sol5, out5] = sdf_nls(model, 'Display', 100, 'CGMaxIter', 100, 'TolFun', 1e-11);
toc

%% Stop here, as the code ahead neads a bug fix
break;

%%
disp('** Compressed tensors - hard coupling - nlsb solver')

model = struct;
model.variables = cpd_rnd([size_one, size_two(1:2)], R, 'real', @rand);
model.transform = 1:5;
model.factorizations{1}.data = {Uc1, Sc1};
model.factorizations{1}.cpd = 1:3;
model.factorizations{1}.weight = 1/snr1;
model.factorizations{2}.data = {Uc2, Sc2};
model.factorizations{2}.cpd = [4 5 3];
model.factorizations{2}.weight = 1/snr2;

lb = cpd_rnd([size_one, size_two(1:2)], R, 'real', @zeros);
ub = cpd_rnd([size_one, size_two(1:2)], R, 'real', @(m,n) ones(m,n)*100);
nlsb = @(F,dF,z0,options) nlsb_gndl(F,dF,lb,ub,z0,options);
tic
[sol5, out5] = sdf_nls(model, 'Display', 10, 'CGMaxIter', 50, 'Algorithm', nlsb);
toc
cpderr(U1, sol5.factors(1:3))
cpderr(U2, sol5.factors([4 5 3]))

%%
disp('** Compressed tensors - flexible coupling - nlsb solver')

model = struct;
model.variables = cpd_rnd([size_one, size_two], R, 'real', @rand);
model.transform = 1:6;
model.factors{8} = [eye(R), -eye(R)];
model.factors{7} = {3, 6};
model.factorizations{1}.data = {Uc1, Sc1};
model.factorizations{1}.cpd = 1:3;
model.factorizations{1}.weight = 1/snr1;
model.factorizations{2}.data = {Uc2, Sc2};
model.factorizations{2}.cpd = 4:6;
model.factorizations{2}.weight = 1/snr2;
model.factorizations{3}.data = zeros(size_one(1),R);
model.factorizations{3}.cpd = [7 8];
model.factorizations{3}.weight = 1/snrC;

lb = cpd_rnd([size_one, size_two], R, 'real', @zeros);
ub = cpd_rnd([size_one, size_two], R, 'real', @(m,n) ones(m,n)*100);
nlsb = @(F,dF,z0,options) nlsb_gndl(F,dF,lb,ub,z0,options);
tic
[sol6, out6] = sdf_nls(model, 'Display', 10, 'CGMaxIter', 50, 'Algorithm', ...
                       nlsb, 'TolFun', 1e-9);
toc
cpderr(U1, sol6.factors(1:3))
cpderr(U2, sol6.factors(4:6))

%% Processing
close all

A = sol4.factors{1};A = A.*repmat(1./sqrt(sum(A.^2)),dim_X(1),1);
B = sol4.factors{2};B = B.*repmat(1./sqrt(sum(B.^2)),dim_X(2),1);
C = sol4.factors{3};C = C.*repmat(1./sqrt(sum(C.^2)),dim_X(3),1);
figure
plot(A)
figure
plot(B)
figure
subplot(211)
plot(C)


D = sol4.factors{4};D = D.*repmat(1./sqrt(sum(D.^2)),dim_Y(1),1);
E = sol4.factors{5};E = E.*repmat(1./sqrt(sum(E.^2)),dim_Y(2),1);
F = sol4.factors{6};F = F.*repmat(1./sqrt(sum(F.^2)),dim_Y(3),1);
figure
plot(D)
figure
plot(E)
figure(3)
subplot(212)
plot(F)

%% Preprocessing for partial

close all

A = sol5.factors.A;A = A.*repmat(1./sqrt(sum(A.^2)),dim_X(1),1);
B = sol5.factors.B;B = B.*repmat(1./sqrt(sum(B.^2)),dim_X(2),1);
C = sol5.factors.C1;C = C.*repmat(1./sqrt(sum(C.^2)),dim_X(3),1);
figure
plot(A)
figure
plot(B)
figure
subplot(211)
plot(C)


D = sol5.factors.D;D = D.*repmat(1./sqrt(sum(D.^2)),dim_Y(1),1);
E = sol5.factors.E;E = E.*repmat(1./sqrt(sum(E.^2)),dim_Y(2),1);
F = sol5.factors.C2;F = F.*repmat(1./sqrt(sum(F.^2)),dim_Y(3),1);
figure
plot(D)
figure
plot(E)
figure(3)
subplot(212)
plot(F)


%% Create tensors 
U1 = cpd_rnd(size_one, R, 'real', @rand);
U2 = cpd_rnd(size_two, R, 'real', @rand);
U2{3} = noisy(U1{3}, 40);
U2{3}(U2{3}<0) = 0;

T1 = cpdgen(U1);
T2 = cpdgen(U2);
T1n = noisy(T1, snr1);
T2n = noisy(T2, snr2);

